@file:OptIn(ExperimentalMaterial3Api::class)

package com.example.app.screens

import android.content.Context
import android.widget.Toast
import androidx.compose.foundation.layout.Column
import androidx.compose.foundation.layout.fillMaxSize
import androidx.compose.foundation.layout.fillMaxWidth
import androidx.compose.foundation.layout.padding
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.filled.Search
import androidx.compose.material3.ExperimentalMaterial3Api
import androidx.compose.material3.Icon
import androidx.compose.material3.Scaffold
import androidx.compose.material3.SearchBar
import androidx.compose.material3.Text
import androidx.compose.material3.TopAppBar
import androidx.compose.runtime.Composable
import androidx.compose.runtime.getValue
import androidx.compose.runtime.mutableStateOf
import androidx.compose.runtime.remember
import androidx.compose.runtime.rememberCoroutineScope
import androidx.compose.runtime.setValue
import androidx.compose.ui.Modifier
import androidx.compose.ui.platform.LocalContext
import androidx.compose.ui.platform.LocalFocusManager
import androidx.compose.ui.res.stringResource
import com.arcgismaps.Color
import com.arcgismaps.geometry.SpatialReference
import com.arcgismaps.mapping.ArcGISMap
import com.arcgismaps.mapping.BasemapStyle
import com.arcgismaps.mapping.Viewpoint
import com.arcgismaps.mapping.symbology.HorizontalAlignment
import com.arcgismaps.mapping.symbology.SimpleMarkerSymbol
import com.arcgismaps.mapping.symbology.SimpleMarkerSymbolStyle
import com.arcgismaps.mapping.symbology.TextSymbol
import com.arcgismaps.mapping.symbology.VerticalAlignment
import com.arcgismaps.mapping.view.Graphic
import com.arcgismaps.mapping.view.GraphicsOverlay
import com.arcgismaps.tasks.geocode.GeocodeParameters
import com.arcgismaps.tasks.geocode.GeocodeResult
import com.arcgismaps.tasks.geocode.LocatorTask
import com.arcgismaps.toolkit.geoviewcompose.MapView
import com.arcgismaps.toolkit.geoviewcompose.MapViewProxy
import com.example.app.R
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Job
import kotlinx.coroutines.launch

@Composable
fun MainScreen() {

    val context = LocalContext.current
    val coroutineScope = rememberCoroutineScope()
    // The focus manager is used to dismiss keyboard after search query submission.
    val focusManager = LocalFocusManager.current
    var queryText by remember { mutableStateOf("") }
    val currentJob = remember { mutableStateOf<Job?>(null) }
    val graphicsOverlay = remember { GraphicsOverlay() }
    val graphicsOverlays = remember { listOf(graphicsOverlay) }
    val mapViewProxy = remember { MapViewProxy() }
    val currentSpatialReference = remember { mutableStateOf<SpatialReference?>(null) }

    val map = remember {
        createMap()
    }

    Scaffold(
        topBar = { TopAppBar(title = { Text(text = stringResource(id = R.string.app_name)) }) }
    ) {

        Column(
            modifier = Modifier
                .fillMaxSize()
                .padding(it)
        ) {

            SearchBar(
                modifier = Modifier.fillMaxWidth(),
                query = queryText,
                onQueryChange = { query -> queryText = query },
                onSearch = { currentQuery ->
                    focusManager.clearFocus()
                    // Cancel any previous search job.
                    currentJob.value?.cancel()
                    // Start a new search job.
                    currentJob.value = coroutineScope.launch {
                        searchAddress(
                            context,
                            coroutineScope,
                            currentQuery,
                            currentSpatialReference.value,
                            graphicsOverlay,
                            mapViewProxy
                        )
                    }
                },
                active = false,
                onActiveChange = { /* Use this for dynamic search results */ },
                leadingIcon = { Icon(Icons.Filled.Search, contentDescription = "Search") },
                placeholder = { Text("Search for an address") }
            ) {}

            MapView(
                modifier = Modifier.fillMaxSize(),
                arcGISMap = map,

                graphicsOverlays = graphicsOverlays,
                mapViewProxy = mapViewProxy,
                onSpatialReferenceChanged = { spatialReference ->
                    currentSpatialReference.value = spatialReference
                }

            )

        }

    }

}

fun createMap(): ArcGISMap {
    return ArcGISMap(BasemapStyle.ArcGISTopographic).apply {
        initialViewpoint = Viewpoint(
            latitude = 34.0270,
            longitude = -118.8050,
            scale = 72000.0
        )
    }
}

suspend fun searchAddress(
    context: Context,
    coroutineScope: CoroutineScope,
    query: String,
    currentSpatialReference: SpatialReference?,
    graphicsOverlay: GraphicsOverlay,
    mapViewProxy: MapViewProxy
) {

    val geocodeServerUri = "https://geocode-api.arcgis.com/arcgis/rest/services/World/GeocodeServer"
    val locatorTask = LocatorTask(geocodeServerUri)

    // Create geocode parameters
    val geocodeParameters = GeocodeParameters().apply {
        resultAttributeNames.add("*")
        maxResults = 1
        outputSpatialReference = currentSpatialReference
    }

    // Search for the address
    locatorTask.geocode(searchText = query, parameters = geocodeParameters)
        .onSuccess { geocodeResults: List<GeocodeResult> ->
            handleGeocodeResults(
                context, coroutineScope, geocodeResults, graphicsOverlay, mapViewProxy
            )
        }.onFailure { error ->
            showMessage(context, "The locatorTask.geocode() call failed: ${error.message}")
        }

}

fun handleGeocodeResults(
    context: Context,
    coroutineScope: CoroutineScope,
    geocodeResults: List<GeocodeResult>,
    graphicsOverlay: GraphicsOverlay,
    mapViewProxy: MapViewProxy
) {

    if (geocodeResults.isNotEmpty()) {
        val geocodeResult = geocodeResults[0]
        // Create a Text graphic to display the address text, and add it to the graphics overlay.
        val textGraphic = createTextGraphic(geocodeResult)
        // Create a red square marker graphic, and add it to the graphics overlay.
        val markerGraphic = createMarkerGraphic(geocodeResult)
        // Clear previous results and add graphics.
        graphicsOverlay.graphics.apply {
            clear()
            add(textGraphic)
            add(markerGraphic)
        }

        coroutineScope.launch {
            val centerPoint = geocodeResult.displayLocation
                ?: return@launch showMessage(context, "The locatorTask.geocode() call failed")

            // Animate the map view to the center point.
            mapViewProxy.setViewpointCenter(centerPoint)
                .onFailure { error ->
                    showMessage(context, "Failed to set Viewpoint center: ${error.message}")
                }

        }

    } else {
        showMessage(context, "No address found for the given query")
    }

}

fun createTextGraphic(geocodeResult: GeocodeResult): Graphic {
    val textSymbol = TextSymbol(
        text = geocodeResult.label,
        color = Color.black,
        size = 18f,
        horizontalAlignment = HorizontalAlignment.Center,
        verticalAlignment = VerticalAlignment.Bottom
    ).apply {
        offsetY = 8f
        haloColor = Color.white
        haloWidth = 2f
    }
    return Graphic(
        geometry = geocodeResult.displayLocation,
        symbol = textSymbol
    )
}

fun createMarkerGraphic(geocodeResult: GeocodeResult): Graphic {
    val simpleMarkerSymbol = SimpleMarkerSymbol(
        style = SimpleMarkerSymbolStyle.Square,
        color = Color.red,
        size = 12.0f
    )
    return Graphic(
        geometry = geocodeResult.displayLocation,
        attributes = geocodeResult.attributes,
        symbol = simpleMarkerSymbol
    )
}

fun showMessage(context: Context, message: String) {
    Toast.makeText(context, message, Toast.LENGTH_LONG).show()
}

